Project(env::Environment, prms) = Project(env, env.mesh, prms)


function Project(env::Environment, mesh::Mesh1d, prms)

  av_derivative_scheme = Symbol(prms["ScalarEq"]["av_derivative_scheme"])
  av_drag = prms["ScalarEq"]["av_drag"]
  hrsc = Symbol(prms["ScalarEq"]["hrsc"])
  slope_limiter_method = Symbol(prms["ScalarEq"]["slope_limiter_method"])
  slope_limiter_tvb_M = prms["ScalarEq"]["slope_limiter_tvb_M"]
  av_recompute_substeps = prms["ScalarEq"]["av_recompute_substeps"]
  muscl_omega = prms["ScalarEq"]["muscl_omega"]
  bernstein = BernsteinReconstruction(mesh)
  analyze_error = prms["ScalarEq"]["analyze_error"]
  analyze_error_norm = prms["ScalarEq"]["analyze_error_norm"]
  fixedprms = (; bernstein, av_derivative_scheme, av_drag, hrsc, muscl_omega,
                 slope_limiter_method, slope_limiter_tvb_M,
                 av_recompute_substeps,
                 analyze_error, analyze_error_norm)

  # construct Project
  equation = make_Equation(prms["ScalarEq"], dimension(mesh))
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])

  # TODO remove this
  bdryconds, ldg_bdryconds, av_bdryconds = nothing, nothing, nothing

  P = Project(equation, hrsc, tci, bdryconds, fixedprms, Dict(prms))

  # register variables
  register!(mesh, P)
  display(mesh.cache)

  # setup initial data
  initialdata!(env, P, prms["ScalarEq"])

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup callbacks
  cbs_step, cbs_substep = make_callback(env, P, isperiodic(mesh))
  append!(env.callbacks, cbs_step)
  append!(env.callbacks_substep, cbs_substep)

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, mesh::Mesh2d, prms)

  bdrycond = dg1d.DirichletBC2()
  analyze_error = prms["ScalarEq"]["analyze_error"]
  analyze_error_norm = prms["ScalarEq"]["analyze_error_norm"]
  fixedprms = (; analyze_error, analyze_error_norm)

  # construct Project
  equation = make_Equation(prms["ScalarEq"], dimension(mesh))
  tci  = TCI.make_TCI(mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(mesh, prms["HRSC"])

  P = Project(equation, hrsc, tci, bdrycond, fixedprms, Dict(prms))

  # register variables
  # TODO Somehow replace register! with register_variables!
  register!(mesh, P)
  # TODO Integrate this into register! dispatch
  _register_bdryconds!(mesh, bdrycond)
  display(mesh.cache)

  # setup initial data
  initialdata!(env, P, prms["ScalarEq"])

  # # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  # setup callbacks
  cbs_step, cbs_substep = make_callback(env, P, isperiodic(mesh))
  append!(env.callbacks, cbs_step)
  append!(env.callbacks_substep, cbs_substep)

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


function register!(mesh::Mesh, P::Project)
  register_equation!(mesh, P)
  register_hrsc!(mesh, P.hrsc)
  register_tci!(mesh, P.tci)
  register_analyze_error!(mesh, P.prms)
end


#######################################################################
#                         register_equation!                          #
#######################################################################


function register_equation!(mesh::Mesh1d, P::Project)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_u, :src_u, :E, :Em1, :F, :Fm1, :v),
    cell_variablenames    = (:max_v,),
    bdry_variablenames    = (:nflx_u,:bdry_u),
    global_variablenames  = (:t, :tm1))
  if mesh.element isa FVElement
    if P.prms.hrsc === :fv_central
      register_variables!(mesh.cache, static_variablenames=(:δu,))
    elseif P.prms.hrsc === :slope_limiter
      register_variables!(mesh.cache,
        static_variablenames=(:u1,:u2,:uL,:uR,:du₊,:du₋,:nflx_u_L, :nflx_u_R,))
    else
      TODO(P.prms.hrsc)
    end
  end
end
function register_equation!(mesh::Mesh2d, P::Project)
  register_variables!(mesh.cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_x, :src_u, :flx_y, :v, :v_x, :v_y),
    cell_variablenames    = (:max_v,),
    bdry_variablenames    = (:nflx_u, :bdry_u),
    global_variablenames  = (:t, :tm1))
end


#######################################################################
#                           register_hrsc!                            #
#######################################################################


function register_hrsc!(mesh, hrsc::Maybe{HRSC.AbstractHRSC}) end


function register_hrsc!(mesh::Mesh1d, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u,),
    bdry_variablenames    = (:bdry_ldg_u, :bdry_smoothed_mu),
    cell_variablenames    = (:mu,:mu_m1,))
end
function register_hrsc!(mesh::Mesh2d, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u_x, :ldg_u_y),
    bdry_variablenames    = (:bdry_ldg_u_x, :bdry_ldg_u_y, :bdry_smoothed_mu,
                             :nflx_u_x, :nflx_u_y),
    cell_variablenames    = (:mu,))
end


function register_hrsc!(mesh::Mesh2d, hrsc::HRSC.EntropyViscosity)
  register_variables!(mesh.cache,
    static_variablenames  = (:ldg_u_x, :ldg_u_y, :E, :Em1, :F_x, :F_y, :F_xm1, :F_ym1),
    bdry_variablenames    = (:bdry_ldg_u_x, :bdry_ldg_u_y, :bdry_smoothed_mu,
                             :nflx_u_x, :nflx_u_y),
    cell_variablenames    = (:mu,))
end


function register_hrsc!(mesh::Mesh1d, hrsc::HRSC.SmoothedArtificialViscosity)
  register_hrsc!(mesh, hrsc.av)
  register_variables!(mesh.cache, static_variablenames  = (:smoothed_mu,))
end
function register_hrsc!(mesh::Mesh2d, hrsc::HRSC.SmoothedArtificialViscosity)
  register_hrsc!(mesh, hrsc.av)
  register_variables!(mesh.cache, static_variablenames  = (:smoothed_mu,))
end


#######################################################################
#                            register_tci!                            #
#######################################################################


function register_tci!(mesh, tci::Nothing) end


function register_tci!(mesh, tci::TCI.AbstractTCI)
  register_variables!(mesh.cache, cell_variablenames = (:flag,))
end


function register_tci!(mesh::Mesh1d, tci::TCI.EntropyProduction)
  register_variables!(mesh.cache,
                      cell_variablenames   = (:flag,),
                      static_variablenames = (:EP,),
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs))
end


########################################################################
#                       register_analyze_error!                        #
########################################################################


function register_analyze_error!(mesh::Mesh, prms)
  register_variables!(mesh.cache, global_variablenames=(:u_norm2,))
  if get(prms, :analyze_error, false)
    register_variables!(mesh.cache, static_variablenames = (:u_err, :u_ana))
    if get(prms, :analyze_error_norm, false)
      register_variables!(mesh.cache, global_variablenames = (:u_err_norm,))
    end
  end
end


#######################################################################
#                        _registere_bdryconds!                         #
#######################################################################


function _register_bdryconds!(mesh, bdryconds) end
function _register_bdryconds!(mesh::Mesh2d, bdryconds::dg1d.AbstractBC)
  register_variables!(mesh.cache,
    static_variablenames  = (:init_u, :init_flx_x, :init_flx_y, :init_v)
  )
end
